(function() {
  var Adapter, FBA_APP, FBA_FIRESTORE, FlameError, all, first, get, getAuth, isEmpty, isEqual, isFunction, isInteger, isPlainObject, isString, last, map, max, min, pick, reverse;

  FBA_APP = require('firebase-admin/app');

  FBA_FIRESTORE = require('firebase-admin/firestore');

  ({getAuth} = require('firebase-admin/auth'));

  first = require('lodash/first');

  FlameError = require('./flame-error');

  get = require('lodash/get');

  isEmpty = require('lodash/isEmpty');

  isEqual = require('lodash/isEqual');

  isFunction = require('lodash/isFunction');

  isInteger = require('lodash/isInteger');

  isPlainObject = require('lodash/isPlainObject');

  isString = require('lodash/isString');

  last = require('lodash/last');

  map = require('lodash/map');

  max = require('lodash/max');

  min = require('lodash/min');

  pick = require('lodash/pick');

  reverse = require('lodash/reverse');

  ({all} = require('rsvp'));

  Adapter = (function() {
    class Adapter {
      constructor(sa = null, name = 'flame-odm') {
        this.cfg = (isString(sa)) ? {} : {sa};
        this.name = name;
        this.cloud = (function() {
          switch (false) {
            case !(isFunction(sa)):
              return 'other';
            case !(isPlainObject(sa)):
              return 'other';
            case !(sa === 'firebase-function'):
              return 'firebase-function';
            case !(sa === 'google-cloud'):
              return 'google-cloud';
            case !(sa === 'process-env'):
              return 'process-env';
            case !(sa === null):
              return 'firebase-function';
            case !(isString(sa)):
              this.name = sa;
              return 'firebase-function';
            default:
              throw new FlameError("Invalid arguments passed to `new Adatper()`.");
          }
        }).call(this);
        return;
      }

      async connect() {
        if (this.connecting !== null) {
          await this.connecting;
          return;
        }
        this.connecting = ((async() => {
          var e, sa;
          if (!(isEmpty(this.fba)) && !(isEmpty(this.db))) {
            return;
          }
          if (this.cloud === 'firebase-function') {
            try {
              FBA_APP.initializeApp();
              this.fba = FBA_APP.getApp();
              this.db = FBA_FIRESTORE.getFirestore(this.fba);
              this.auth = getAuth(this.fba);
            } catch (error) {
              e = error;
              throw new FlameError("There was an error connecting to Firebase. Ref: 'firebase-function'");
              console.log(e);
            }
            return;
          }
          if (this.cloud === 'google-cloud') {
            try {
              FBA_APP.initializeApp({
                credential: FBA_APP.applicationDefault()
              });
              this.fba = FBA_APP.getApp();
              this.db = FBA_FIRESTORE.getFirestore(this.fba);
              this.auth = getAuth(this.fba);
            } catch (error) {
              e = error;
              throw new FlameError("There was an error connecting to Firebase. Ref: 'google-cloud'");
              console.log(e);
            }
            return;
          }
          if (this.cloud === 'process-env') {
            try {
              sa = JSON.parse(process.env.FB_SERVICE_ACCOUNT);
              FBA_APP.initializeApp({
                credential: FBA_APP.cert(sa)
              }, this.name);
              this.fba = FBA_APP.getApp(this.name);
              this.db = FBA_FIRESTORE.getFirestore(this.fba);
              this.auth = getAuth(this.fba);
            } catch (error) {
              e = error;
              throw new FlameError("There was an error connecting to Firebase. Ref: 'other'");
              console.log(e);
            }
            return;
          }
          if (this.cloud === 'other') {
            try {
              sa = this.cfg.sa;
              if (isFunction(sa)) {
                sa = (await sa());
              }
              FBA_APP.initializeApp({
                credential: FBA_APP.cert(sa)
              }, this.name);
              this.fba = FBA_APP.getApp(this.name);
              this.db = FBA_FIRESTORE.getFirestore(this.fba);
              this.auth = getAuth(this.fba);
            } catch (error) {
              e = error;
              throw new FlameError("There was an error connecting to Firebase. Ref: 'other'");
              console.log(e);
            }
          }
        })());
        return (await this.connecting);
      }

      async count(model, query) {
        var col_ref, count, err, fbq;
        await this.connect();
        col_ref = this.db.collection(`${model.collection}`);
        fbq = query.prepare(col_ref, model.serializer);
        try {
          count = (await fbq.count().get());
          if (isInteger(count != null ? typeof count.data === "function" ? count.data().count : void 0 : void 0)) {
            return count.data().count;
          }
        } catch (error) {
          err = error;
          console.log(err);
          ((function() {})());
        }
        return null;
      }

      async del(model, id, transaction) {
        return (await this.connect());
      }

      async destroy(record, transaction = null) {
        var dr, err;
        await this.connect();
        dr = this.db.doc(`${record.collection}/${record.id}`);
        if (transaction) {
          await (transaction.delete(dr));
        } else {
          try {
            await dr.delete();
          } catch (error) {
            err = error;
            console.log(err);
            return false;
          }
        }
        return true;
      }

      async find(model, query, fields = null, transaction = null) {
        var ls, ref;
        await this.connect();
        ls = (await (this.findAll(model, query, fields, transaction)));
        return (ref = first(ls)) != null ? ref : null;
      }

      async get(model, id, fields = null, transaction = null) {
        var dr, ds, err, obj;
        await this.connect();
        dr = this.db.doc(`${model.collection}/${id}`);
        if (transaction) {
          ds = (await (transaction.get(dr)));
          if (ds.exists) {
            obj = model.serializer.fromDB(ds.data());
            if (fields) {
              return pick(obj, fields);
            } else {
              return obj;
            }
          }
          return null;
        } else {
          try {
            ds = (await dr.get());
            if (ds.exists) {
              obj = model.serializer.fromDB(ds.data());
              if (fields) {
                return pick(obj, fields);
              } else {
                return obj;
              }
            }
          } catch (error) {
            err = error;
            console.log(err);
            ((function() {})());
          }
        }
        return null;
      }

      async getAll(model, ids, fields = null) {
        var drs, dss;
        await this.connect();
        drs = map(ids, ((id) => {
          return this.db.doc(`${model.collection}/${id}`);
        }));
        dss = (await (this.db.getAll(...drs)));
        if (!(isEmpty(dss))) {
          return map(dss, (function(ds) {
            var obj;
            obj = model.serializer.fromDB(ds.data());
            if (fields) {
              return pick(obj, fields);
            } else {
              return obj;
            }
          }));
        }
        return null;
      }

      async findAll(model, query, fields = null, transaction = null) {
        var col_ref, err, fbq, fs, qs, s;
        await this.connect();
        col_ref = this.db.collection(`${model.collection}`);
        fbq = query.prepare(col_ref, model.serializer);
        if (fields) {
          s = model.serializer;
          fs = map(fields, function(f) {
            return s.fmt[s.fmts.field.db](f);
          });
          fbq = fbq.select(...fs);
        }
        if (fields === []) {
          fbq = fbq.select();
        }
        if (transaction) {
          qs = (await (transaction.get(fbq)));
          if (!qs.empty) {
            return map(qs.docs, (function(ds) {
              return model.serializer.fromDB(ds.data());
            }));
          }
          return null;
        } else {
          try {
            qs = (await fbq.get());
            if (!qs.empty) {
              return map(qs.docs, (function(ds) {
                return model.serializer.fromDB(ds.data());
              }));
            }
          } catch (error) {
            err = error;
            console.log(err);
            ((function() {})());
          }
          return null;
        }
      }

      async page(model, pager = null, cursor = null, fields = null, transaction = null, counts = true) {
        var CPEND, after, at_end, at_start, before, coll_first, coll_last, items, itemz, next, pg_count, previous, priorz, queries, ref, ref1, ref2, ref3, tail_count, total;
        CPEND = (get(cursor, 'position')) === 'page-end';
        await this.connect();
        queries = pager.queries(cursor);
        [coll_first, coll_last, itemz, priorz, total, tail_count] = (await (all([this.find(model, queries.collection, fields, transaction), this.find(model, queries.reversed, fields, transaction), this.findAll(model, queries.itemz, fields, transaction), this.findAll(model, queries.priorz, fields, transaction), (counts ? this.count(model, queries.collection) : 0), (counts ? this.count(model, queries.tail) : 0)])));
        pg_count = min([pager.size, itemz.length]);
        items = itemz.slice(0, pg_count);
        if (CPEND) {
          items = reverse(items);
          [coll_first, coll_last] = [coll_last, coll_first];
        }
        at_start = isEqual(coll_first, first(items));
        at_end = isEqual(coll_last, last(items));
        previous = (ref = (!at_start ? {
          obj: priorz[1],
          position: 'page-end'
        } : void 0)) != null ? ref : null;
        next = (ref1 = (!at_end ? {
          obj: itemz[pg_count],
          position: 'page-start'
        } : void 0)) != null ? ref1 : null;
        after = tail_count - pg_count;
        before = total - tail_count;
        if (CPEND) {
          [after, before] = [before, after];
          previous = (ref2 = (!at_start ? {
            obj: itemz[pg_count],
            position: 'page-end'
          } : void 0)) != null ? ref2 : null;
          next = (ref3 = (!at_end ? {
            obj: priorz[1],
            position: 'page-start'
          } : void 0)) != null ? ref3 : null;
        }
        return {
          counts: {
            total: counts ? total : null,
            before: counts ? before : null,
            page: counts ? pg_count : null,
            after: counts ? after : null
          },
          collection: {
            first: coll_first,
            last: coll_last
          },
          page: {
            first: first(items),
            items: items,
            last: last(items)
          },
          cursors: {
            previous: previous,
            current: cursor,
            next: next
          }
        };
      }

      async save(record, transaction = null) {
        var dr, err;
        await this.connect();
        dr = this.db.doc(`${record.collection}/${record.id}`);
        if (transaction) {
          await (transaction.create(dr, record.serializer.toDB(record.data)));
        } else {
          try {
            await (dr.create(record.serializer.toDB(record.data)));
          } catch (error) {
            err = error;
            console.log(err);
            return false;
          }
        }
        return true;
      }

      async transact(fn) {
        var err;
        await this.connect();
        try {
          return (await (this.db.runTransaction((async function(t) {
            return (await (fn(t)));
          }))));
        } catch (error) {
          err = error;
          console.log(err);
          return null;
        }
      }

      async traverse(model, pager = null, fn, fields = null) {
        var cursor, items, pg;
        cursor = null;
        pg = null;
        while (!pg || cursor) {
          pg = (await (this.page(model, pager, cursor, fields, null, false)));
          items = get(pg, 'page.items', []);
          cursor = get(pg, 'cursors.next');
          await (all(map(items, (async function(r) {
            return (await (fn(r)));
          }))));
        }
      }

      async update(record, fields = [], transaction = null) {
        var dr, err;
        await this.connect();
        dr = this.db.doc(`${record.collection}/${record.id}`);
        if (transaction) {
          await (transaction.update(dr, record.serializer.toDB(pick(record.data, fields))));
        } else {
          try {
            await (dr.update(record.serializer.toDB(pick(record.data, fields))));
          } catch (error) {
            err = error;
            console.log(err);
            return false;
          }
        }
        return true;
      }

    };

    Adapter.prototype.type = 'Adapter';

    Adapter.prototype.connecting = null;

    return Adapter;

  }).call(this);

  module.exports = Adapter;

}).call(this);
