(function() {
  var FlameError, Pager, Query, Serializer, cloneDeep, each, first, get, includes, isInteger, join, last, map, reduce, union;

  cloneDeep = require('lodash/cloneDeep');

  each = require('lodash/each');

  first = require('lodash/first');

  get = require('lodash/get');

  includes = require('lodash/includes');

  isInteger = require('lodash/isInteger');

  join = require('lodash/join');

  last = require('lodash/last');

  map = require('lodash/map');

  reduce = require('lodash/reduce');

  union = require('lodash/union');

  Query = require('./query');

  FlameError = require('./flame-error');

  Serializer = require('./serializer');

  Pager = (function() {
    var allowed;

    class Pager {
      constructor(q = [], opts = {}) {
        var e, ref;
        if (!(reduce(q, (function(out, constraint) {
          return out && (includes(allowed, constraint[0]));
        }), true))) {
          e = `The first argument to a Pager may only use the following constraints: ${join(allowed, ', ')}.`;
          throw new FlameError(e);
          return;
        }
        this.q = q;
        this.size = (ref = opts.size) != null ? ref : 10;
        return;
      }

      queries(cursor = null) {
        var collection_q, itemz_q, priorz_q, reversed_q, start_at_q, tail_q;
        start_at_q = (() => {
          var order_fields, order_values;
          if (cursor) {
            order_fields = reduce(this.q, (function(acc, _q) {
              if (_q[0] === 'order-by') {
                return union(acc, [_q[1]]);
              } else {
                return acc;
              }
            }), []);
            order_values = map(order_fields, (function(_f) {
              return get(cursor.obj, _f);
            }));
            return [['start-at', order_values]];
          } else {
            return [];
          }
        })();
        collection_q = cloneDeep(this.q);
        reversed_q = map(cloneDeep(collection_q), (function(_q) {
          switch (false) {
            case !(((first(_q)) === 'order-by') && ((last(_q)) === 'desc')):
              return [..._q.slice(0, 2), 'asc'];
            case !(((first(_q)) === 'order-by') && ((last(_q)) === 'asc')):
              return [..._q.slice(0, 2), 'desc'];
            case !(((first(_q)) === 'order-by') && (_q.length === 2)):
              return [..._q.slice(0, 2), 'desc'];
            default:
              return _q;
          }
        }));
        if (cursor && cursor.position === 'page-end') {
          [collection_q, reversed_q] = [reversed_q, collection_q];
        }
        itemz_q = union(cloneDeep(collection_q), [['limit', this.size + 1]], start_at_q);
        priorz_q = union(cloneDeep(reversed_q), [['limit', 2]], start_at_q);
        tail_q = union(cloneDeep(collection_q), start_at_q);
        return {
          collection: new Query(collection_q),
          reversed: new Query(reversed_q),
          itemz: new Query(itemz_q),
          priorz: new Query(priorz_q),
          tail: new Query(tail_q)
        };
      }

    };

    Pager.prototype.type = 'Pager';

    allowed = ['eq', 'eq-any', 'gt', 'gte', 'includes', 'includes-any', 'lt', 'lte', 'not-eq', 'not-eq-any', 'order-by', 'page-size'];

    return Pager;

  }).call(this);

  module.exports = Pager;

}).call(this);
