(function() {
  var Query, Serializer, each;

  each = require('lodash/each');

  Serializer = require('./serializer');

  Query = (function() {
    var ops;

    class Query {
      constructor(q = [], serializer = null) {
        this.serializer = serializer != null ? serializer : new Serializer();
        this.q = q;
        return;
      }

      prepare(col_ref, serializer = null) {
        var fbq, s;
        s = serializer != null ? serializer : this.serializer;
        fbq = col_ref;
        each(this.q, function(statement) {
          fbq = ops[statement[0]](fbq, s, ...statement.slice(1));
        });
        return fbq;
      }

    };

    Query.prototype.type = 'Query';

    // fbq = firestore query, s = serializer, fs = fields, v(s) = value(s)
    ops = {
      'eq': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), '==', v);
      },
      'eq-any': function(fbq, s, fs, vs) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), 'in', vs);
      },
      'gt': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), '>', v);
      },
      'gte': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), '>=', v);
      },
      'includes': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), 'array-contains', v);
      },
      'includes-any': function(fbq, s, fs, vs) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), 'array-contains-any', vs);
      },
      'limit': function(fbq, _, n) {
        return fbq.limit(n);
      },
      'lt': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), '<', v);
      },
      'lte': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), '<=', v);
      },
      'not-eq': function(fbq, s, fs, v) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), '!=', v);
      },
      'not-eq-any': function(fbq, s, fs, vs) {
        return fbq.where(s.fmt[s.fmts.field.db](fs), 'not-in', vs);
      },
      'order-by': function(fbq, s, fs, d) {
        return fbq.orderBy(s.fmt[s.fmts.field.db](fs), d != null ? d : 'asc');
      },
      'start-at': function(fbq, _, v) {
        return fbq.startAt(...v);
      }
    };

    return Query;

  }).call(this);

  module.exports = Query;

}).call(this);
